from typing import Sequence

import numpy as np
import pandas as pd
import pandas.api.types as pdt

import visions
from visions.relations import IdentityRelation, TypeRelation
from visions.typesets.typeset import get_type_from_path


class Nominal(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Categorical)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return not pdt.is_categorical_dtype(series) or (
            pdt.is_categorical_dtype(series) and not series.cat.ordered
        )


class Categorical(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        # This example could be extended to show how low-cardinality discrete variables would be
        # inferred to nominal / ordinal. This can be achieved with an InferenceRelation.
        return [IdentityRelation(visions.Generic)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_object_dtype(series)


class Binary(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Nominal)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        state["n_distinct"] = state.get("n_distinct") or series.nunique()
        return state["n_distinct"] == 2


class Ordinal(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Categorical)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_categorical_dtype(series) and series.cat.ordered


class Numeric(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(visions.Generic)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_numeric_dtype(series)


class Continuous(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Numeric)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return not pdt.is_integer_dtype(series)


class Discrete(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Numeric)]

    @classmethod
    def contains_op(cls, series: pd.Series, state: dict) -> bool:
        return pdt.is_integer_dtype(series)


class VariableTypeset(visions.VisionsTypeset):
    def __init__(self):
        types = {
            visions.Generic,
            Categorical,
            Nominal,
            Ordinal,
            Numeric,
            Continuous,
            Discrete,
            Binary,
        }
        super().__init__(types)


variable_set = VariableTypeset()
variable_set.output_graph("variable_set.pdf")


class Classification(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(visions.Generic)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        return state["dtype"] in [Nominal, Categorical, Ordinal, Binary]


class BinaryClassification(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Classification)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        return state["dtype"] == Binary


class MultiClassification(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Classification)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        return state["dtype"] != Binary


class Regression(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(visions.Generic)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        return state["dtype"] in [Continuous, Discrete]


class PoissonRegression(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Regression)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        if not state["dtype"] == Discrete:
            return False

        # This is a simplified test if poisson regression applies that doesn't take into account if
        # the ratio is significant
        state["mean_var_ratio"] = state.get("mean_var_rate") or np.mean(
            series
        ) / np.var(series)
        return np.isclose(state["mean_var_ratio"], 1, rtol=0.05)


class NegBinomRegression(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Regression)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        if not state["dtype"] == Discrete:
            return False

        # See comment at poisson regression
        state["mean_var_ratio"] = state.get("mean_var_rate") or np.mean(
            series
        ) / np.var(series)
        return state["mean_var_ratio"] > 1.05


class OrdinalRegression(visions.VisionsBaseType):
    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        return [IdentityRelation(Classification)]

    @classmethod
    def contains_op(cls, series, state):
        state["dtype"] = state.get("dtype") or variable_set.detect_type(series)
        return state["dtype"] == Ordinal


class MLProblemTypeset(visions.VisionsTypeset):
    def __init__(self):
        types = {
            visions.Generic,
            Classification,
            BinaryClassification,
            MultiClassification,
            Regression,
            NegBinomRegression,
            PoissonRegression,
            OrdinalRegression,
        }
        super().__init__(types)


problem_set = MLProblemTypeset()
problem_set.output_graph("problem_set.pdf")


# Example
dataset = pd.DataFrame(
    {
        "target_3": ["cat", "dog", "dog", "cat", "horse"],
        "target_2": ["cat", "dog", "dog", "cat", "dog"],
        "target_num": [1, 2, 2, 1, 2],
    }
)


for target in dataset.columns:
    _, problem_types, state = problem_set.detect(dataset[target])
    problem_type = get_type_from_path(problem_types)

    print(
        f"The target variable '{target}' is of the {state['dtype']} statistical type."
    )
    print(f"Our logic found that a {problem_type} model should be used.")



import sys

import pytest


@pytest.mark.spark_test
def test_import_spark_session(spark_session):
    """
    Test if its possible to import spark
    """
    try:
        import pyspark
        from pyspark import SparkConf, SparkContext
        from pyspark.sql import SparkSession
    except ImportError:
        pytest.fail(
            """Could not import pyspark - is SPARK_HOME and JAVA_HOME set as variables?
                    see https://spark.apache.org/docs/latest/quick-start.html and ensure
                    that your spark instance is configured properly"""
        )


@pytest.mark.spark_test
def test_create_spark_session(spark_session):
    """
    Test if pytest-spark's spark sessions can be properly created
    """
    try:
        from pyspark.sql import SparkSession

        assert isinstance(spark_session, SparkSession)
    except AssertionError:
        pytest.fail(
            """pytest spark_session was not configured properly and could not be created
        is pytest-spark installed and configured properly?"""
        )


@pytest.fixture()
def spark_df(spark_session):
    from datetime import datetime
    from decimal import Decimal

    from pyspark.sql.types import (
        DateType,
        DecimalType,
        IntegerType,
        StringType,
        StructField,
        StructType,
    )

    data2 = [
        ("James", "", "Smith", "36636", "M", 3000, datetime(1980, 1, 2), Decimal(1.5)),
        ("Michael", "Rose", "", "40288", "M", 4000, datetime(1980, 1, 1), Decimal(1.4)),
        (
            "Robert",
            "",
            "Williams",
            "42114",
            "M",
            4000,
            datetime(1980, 1, 3),
            Decimal(1.2),
        ),
        (
            "Maria",
            "Anne",
            "Jones",
            "39192",
            "F",
            4000,
            datetime(1980, 1, 1),
            Decimal(1.0),
        ),
        ("Jen", "Mary", "Brown", "", "F", -1, datetime(1980, 1, 1), Decimal(0.5)),
    ]

    schema = StructType(
        [
            StructField("firstname", StringType(), True),
            StructField("middlename", StringType(), True),
            StructField("lastname", StringType(), True),
            StructField("id", StringType(), True),
            StructField("gender", StringType(), True),
            StructField("salary", IntegerType(), True),
            StructField("da", DateType(), True),
            StructField("nu", DecimalType(), True),
        ]
    )

    return spark_session.createDataFrame(data=data2, schema=schema)


@pytest.mark.spark_test
def test_spark_backend(spark_session, spark_df):
    import visions
    from visions import Date, Float, Integer, Object, StandardSet, String

    tset = StandardSet()
    result = tset.detect_type(spark_df)
    expected = {
        "da": Object,
        "firstname": String,
        "gender": String,
        "id": String,
        "lastname": String,
        "middlename": String,
        "nu": Float,
        "salary": Integer,
    }
    assert result == expected

    tset2 = tset + Date
    result2 = tset2.detect_type(spark_df)
    expected2 = {
        "da": Date,
        "firstname": String,
        "gender": String,
        "id": String,
        "lastname": String,
        "middlename": String,
        "nu": Float,
        "salary": Integer,
    }
    assert result2 == expected2



import numpy as np

from visions import Generic, VisionsTypeset, create_type


def is_np_float(array, state):
    return np.issubdtype(array.dtype, np.float)


def is_np_int(array, state):
    return np.issubdtype(array.dtype, np.int)


def float_is_int(array, state):
    try:
        return (array.astype(int) == array).all()
    except:
        return False


def to_int(array, state):
    return array.astype(int)


Float = create_type(
    "Float",
    identity=Generic,
    contains=is_np_float,
)

Integer = create_type(
    "Integer",
    identity=Generic,
    contains=is_np_int,
    inference=dict(related_type=Float, relationship=float_is_int, transformer=to_int),
)

my_typeset = VisionsTypeset({Generic, Integer, Float})
my_typeset.output_graph("declarative_set.pdf")



from typing import Dict, Set, Type

import pytest

from visions import StandardSet, VisionsBaseType
from visions.backends.python.sequences import get_sequences
from visions.test.utils import (
    cast,
    contains,
    convert,
    get_cast_cases,
    get_contains_cases,
    get_convert_cases,
    get_inference_cases,
    infers,
)
from visions.types import (
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Generic,
    Integer,
    Object,
    String,
    TimeDelta,
)

sequences = get_sequences()

typeset = StandardSet()

contains_map: Dict[Type[VisionsBaseType], Set[str]] = {
    Integer: {
        "int_series",
        "int_range",
        "int_series_boolean",
    },
    Float: {
        "float_series",
        "float_series2",
    },
    Categorical: set(),
    Boolean: {"bool_series", "bool_nan_series"},
    Complex: {
        "complex_series_py",
        "complex_series_py_float",
    },
    DateTime: set(),
    TimeDelta: set(),
    String: {
        "py_datetime_str",
        "timestamp_string_series",
        "string_series",
        "string_unicode_series",
        "path_series_linux_str",
        "path_series_windows_str",
        "string_date",
        "textual_float",
        "textual_float_nan",
        "ip_str",
        "string_flt",
        "string_num",
        "str_url",
        "string_bool_nan",
        "str_complex",
        "str_complex_nan",
        "uuid_series_str",
        "str_int_leading_zeros",
        "email_address_str",
        "str_float_non_leading_zeros",
        "str_int_zeros",
    },
    Object: {
        "path_series_linux",
        "path_series_linux_missing",
        "path_series_windows",
        "url_series",
        "url_none_series",
        "file_test_py",
        "file_mixed_ext",
        "file_test_py_missing",
        "image_png",
        "image_png_missing",
        "image_png",
        "image_png_missing",
        "email_address",
        "email_address_missing",
        "uuid_series",
        "uuid_series_missing",
        "ip",
        "ip_mixed_v4andv6",
        "ip_missing",
        "mixed_list[str,int]",
        "mixed_dict",
        "callable",
        "module",
        "mixed_integer",
        "mixed_list",
        "date",
        "time",
    },
    Generic: {
        "empty",
    },
}


@pytest.mark.parametrize(**get_contains_cases(sequences, contains_map, typeset))
def test_contains(name, series, contains_type, member):
    """Test the generated combinations for "series in type"

    Args:
        series: the series to test
        contains_type: the type to test against
        member: the result
    """
    result, message = contains(name, series, contains_type, member)
    assert result, message


inference_map = {
    "int_series": Integer,
    "int_range": Integer,
    "float_series": Float,
    "int_series_boolean": Integer,
    "float_series2": Integer,
    "complex_series_float": Integer,
    "string_series": String,
    "categorical_string_series": Categorical,
    "timestamp_string_series": String,
    "string_unicode_series": String,
    "string_num": Integer,
    "complex_series_py_float": Integer,
    "string_flt": Float,
    "string_bool_nan": Boolean,
    "string_date": String,
    "py_datetime_str": DateTime,
    "str_url": String,
    "bool_series": Boolean,
    "bool_nan_series": Boolean,
    "complex_series_py": Complex,
    "geometry_series_missing": Object,
    "geometry_series": Object,
    "path_series_linux": Object,
    "path_series_linux_missing": Object,
    "path_series_linux_str": String,
    "path_series_windows": Object,
    "path_series_windows_str": String,
    "url_series": Object,
    "url_none_series": Object,
    "mixed_list[str,int]": Object,
    "mixed_dict": Object,
    "mixed_integer": Object,
    "mixed_list": Object,
    "callable": Object,
    "module": Object,
    "textual_float": Float,
    "textual_float_nan": Float,
    "empty": Generic,
    "empty_object": Generic,
    "empty_float": Generic,
    "ip": Object,
    "ip_str": String,
    "ip_missing": Object,
    "date": Object,
    "time": Object,
    "str_complex": Complex,
    "str_complex_nan": Complex,
    "uuid_series": Object,
    "uuid_series_str": String,
    "uuid_series_missing": Object,
    "ip_mixed_v4andv6": Object,
    "file_test_py": Object,
    "file_test_py_missing": Object,
    "file_mixed_ext": Object,
    "image_png": Object,
    "image_png_missing": Object,
    "str_int_leading_zeros": String,
    "str_float_non_leading_zeros": Float,
    "str_int_zeros": Integer,
    "email_address": Object,
    "email_address_missing": Object,
    "email_address_str": String,
}


@pytest.mark.parametrize(**get_inference_cases(sequences, inference_map, typeset))
def test_inference(name, series, inference_type, typeset, difference):
    """Test the generated combinations for "inference(series) == type"

    Args:
        series: the series to test
        inference_type: the type to test against
    """
    result, message = infers(name, series, inference_type, typeset, difference)
    assert result, message


# Conversions in one single step
convert_map = [
    # Model type, Relation type
    (Integer, Float, {"float_series2"}),
    (
        Complex,
        String,
        {
            "str_complex",
            "str_complex_nan",
            "textual_float_nan",
            "str_int_zeros",
            "textual_float",
            "string_num",
            "string_flt",
            "str_float_non_leading_zeros",
        },
    ),
    (
        Float,
        String,
        {
            "string_flt",
            "string_num",
            "textual_float",
            "textual_float_nan",
            "str_float_non_leading_zeros",
            "str_int_zeros",
        },
    ),
    (Boolean, String, {"string_bool_nan"}),
    (Boolean, Object, {"bool_nan_series"}),
    (Float, Complex, {"complex_series_py_float"}),
    (DateTime, String, {"py_datetime_str"}),
]


@pytest.mark.parametrize(**get_convert_cases(sequences, convert_map, typeset))
def test_conversion(name, source_type, relation_type, series, member):
    """Test the generated combinations for "convert(series) == type" and "infer(series) = source_type"

    Args:
        series: the series to test
        type: the type to test against
    """
    result, message = convert(name, source_type, relation_type, series, member)
    assert result, message


#
# cast_results = {
#     "float_series2": [1, 2, 3, 4],
#     "string_num": pd.Series([1, 2, 3], dtype=np.int64),
#     "string_flt": pd.Series([1.0, 45.67, 3.5], dtype=np.float64),
#     "string_bool_nan": pd.Series([True, False, None], dtype=hasnan_bool_name),
#     "str_float_non_leading_zeros": pd.Series([0.0, 0.04, 0.0], dtype=np.float64),
#     "str_int_zeros": pd.Series([0, 0, 0, 2], dtype=np.int64),
#     "bool_nan_series": pd.Series([True, False, None], dtype=hasnan_bool_name),
#     "str_complex": pd.Series(
#         [complex(1, 1), complex(2, 2), complex(10, 100)], dtype=np.complex128
#     ),
#     "str_complex_nan": pd.Series(
#         [complex(1, 1), complex(2, 2), complex(10, 100), np.nan], dtype=np.complex128
#     ),
#     "textual_float": pd.Series([1.1, 2.0], dtype=np.float64),
#     "textual_float_nan": pd.Series([1.1, 2.0, np.nan], dtype=np.float64),
#     "mixed": pd.Series([True, False, None], dtype=hasnan_bool_name),
# }
#
#
# @pytest.mark.parametrize(**get_cast_cases(series, cast_results))
# def test_cast(series, expected):
#     if isinstance(expected, str):
#         expected = None
#     result, message = cast(series, typeset, expected)
#     assert result, message



from typing import Dict, Set, Type, cast

import pytest

from visions import StandardSet, VisionsBaseType
from visions.backends.pandas.test_utils import pandas_version
from visions.test.series_sparse import get_sparse_series
from visions.test.utils import contains, get_contains_cases, get_inference_cases, infers
from visions.types import (
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Generic,
    Integer,
    Object,
    String,
    TimeDelta,
)

not_pandas_1_0_5 = not (
    (pandas_version[0] == 1) and (pandas_version[1] == 0) and (pandas_version[2] == 5)
)

series = get_sparse_series()

typeset = StandardSet()

contains_map: Dict[Type[VisionsBaseType], Set[str]] = {
    DateTime: set(),
    TimeDelta: set(),
    Categorical: set(),
    Object: set(),
    Integer: set(),
    Complex: set(),
    Float: set(),
    Boolean: set(),
    String: set(),
    Generic: {
        "int_sparse",
        "pd_int64_sparse",
        "float_sparse",
        "bool_sparse",
        "complex_sparse",
        "str_obj_sparse",
    },
}


if pandas_version[0] >= 1 and not_pandas_1_0_5:
    contains_map[Generic].add("pd_bool_sparse")
    contains_map[Generic].add("pd_string_sparse")


@pytest.mark.parametrize(**get_contains_cases(series, contains_map, typeset))
def test_contains(name, series, contains_type, member):
    """Test the generated combinations for "series in type"

    Args:
        series: the series to test
        contains_type: the type to test against
        member: the result
    """
    result, message = contains(name, series, contains_type, member)
    assert result, message


inference_map: Dict[str, Type[VisionsBaseType]] = {
    "int_sparse": Generic,
    "pd_int64_sparse": Generic,
    "float_sparse": Generic,
    "bool_sparse": Generic,
    "pd_bool_sparse": Generic,
    "complex_sparse": Generic,
    "str_obj_sparse": Generic,
    "pd_categorical_sparse": Generic,
    # "datetime_sparse": Generic,
}

if pandas_version[0] >= 1 and not_pandas_1_0_5:
    inference_map["pd_bool_sparse"] = Generic
    inference_map["pd_string_sparse"] = Generic


@pytest.mark.parametrize(**get_inference_cases(series, inference_map, typeset))
def test_inference(name, series, inference_type, typeset, difference):
    """Test the generated combinations for "inference(series) == type"

    Args:
        series: the series to test
        inference_type: the type to test against
    """
    result, message = infers(name, series, inference_type, typeset, difference)
    assert result, message


@pytest.mark.parametrize("series", series)
def test_detect_type(series):
    typeset.detect_type(series)


@pytest.mark.parametrize("series", series)
def test_cast_inferred(series):
    typeset.cast_to_inferred(series)



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import series_not_empty, series_not_sparse
from visions.types.categorical import Categorical


@Categorical.contains_op.register
@series_not_sparse
@series_not_empty
def categorical_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_categorical_dtype(series)



import numpy as np

from visions import Generic, VisionsTypeset
from visions.declarative import create_type


def is_np_float(array, state):
    return np.issubdtype(array.dtype, np.float)


def is_np_int(array, state):
    return np.issubdtype(array.dtype, np.int)


def float_is_int(array, state):
    try:
        return (array.astype(int) == array).all()
    except:
        return False


def to_int(array, state):
    return array.astype(int)


def test_declarative():
    Float = create_type(
        "Float",
        identity=Generic,
        contains=is_np_float,
    )

    Integer = create_type(
        "Integer",
        identity=Generic,
        contains=is_np_int,
        inference=dict(
            related_type=Float, relationship=float_is_int, transformer=to_int
        ),
    )

    my_typeset = VisionsTypeset({Generic, Integer, Float})

    array = np.arange(1, 100).astype(np.float)
    assert my_typeset.detect_type(array) == Float
    assert my_typeset.infer_type(array) == Integer



import numpy as np

from visions.backends.numpy.types.integer import (
    float_is_integer,
    float_to_integer,
    integer_contains,
)


def test_sequences():
    assert (
        float_to_integer(np.array([1.0, 2.0, 3.0]), {}) == np.array([1, 2, 3])
    ).all()
    assert float_is_integer(np.array([1.0, 2.0, 3.0]), {})
    assert not float_is_integer(np.array([1.2, 2.0, 3.0]), {})
    assert integer_contains(np.array([1, 2, 3]), {})
    assert not integer_contains(np.array([], dtype=np.int8), {})
    assert not integer_contains(np.array([1.2, 2.3]), {})



import copy
import sys

import pytest

from visions.test.series import get_series
from visions.test.series_geometry import get_geometry_series
from visions.test.utils import sequences_equal
from visions.typesets import CompleteSet


def all_relations_tested(series_map):
    typeset = CompleteSet()

    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            "Not all inferential relations are tested {missing_relations}".format(
                missing_relations=missing_relations
            )
        )


def pytest_generate_tests(metafunc):
    _test_suite = get_series()
    _test_suite.update(get_geometry_series())
    if metafunc.function.__name__ in [
        "test_consistency",
        "test_side_effects",
        "test_multiple_inference",
    ]:
        argsvalues = []
        for name, series in _test_suite.items():
            args = {"id": name}
            argsvalues.append(pytest.param(name, series, **args))

        metafunc.parametrize(argnames=["name", "series"], argvalues=argsvalues)


def test_consistency(name, series):
    typeset = CompleteSet()

    if (
        name in ["timedelta_series_nat", "date_series_nat", "timestamp_series_nat"]
        and sys.version_info.major == 3
        and sys.version_info.minor == 7
    ):
        pytest.skip("unsupported configuration")

    initial_type = str(typeset.detect_type(series))
    converted_type = str(typeset.infer_type(series))

    if initial_type != converted_type:
        converted_series = typeset.cast_to_inferred(series.copy(deep=True))

        if hasattr(series, "dtype") and hasattr(converted_series, "dtype"):
            assert (
                series.dtype.kind != converted_series.dtype.kind
                or not sequences_equal(series, converted_series)
            )
        else:
            assert not sequences_equal(series, converted_series)

    else:
        converted_series = typeset.cast_to_inferred(series)
        assert sequences_equal(series, converted_series)


def test_side_effects(name, series):
    reference = series.copy()

    typeset = CompleteSet()
    typeset.detect_type(series)
    typeset.infer_type(series)

    assert sequences_equal(series, reference)


def test_multiple_inference(name, series):
    """
    Notes:
        Copy to prevent possible side effects only for testing.
    """
    ts = CompleteSet()

    inferred_type = str(ts.infer_type(series))

    series_convert = ts.cast_to_inferred(copy.copy(series))

    initial_type_after_convert = str(ts.detect_type(series_convert))
    assert inferred_type == initial_type_after_convert

    series_convert2 = ts.cast_to_inferred(series_convert)

    inferred_type_after_convert = str(ts.infer_type(series_convert2))
    assert initial_type_after_convert == inferred_type_after_convert
    assert sequences_equal(series_convert, series_convert2)



# from visions.relations import IdentityRelation, TypeRelation
# from visions.relations.integer_to_datetime import integer_to_datetime_year_month_day
# from visions.types.date_time import DateTime
# from visions.types.integer import Integer
#
#
# def relations_generator(vtype):
#     return [integer_to_datetime_year_month_day(vtype)]
#
#
# def test_evolve_extend_datetime():
#     DateTimeIntYYYYMMDD = DateTime.evolve_type("int_yyyymmdd", relations_generator)
#     assert str(DateTimeIntYYYYMMDD) == "DateTime[int_yyyymmdd]"
#     assert str(DateTime) == "DateTime"
#     assert len(DateTime.get_relations()) == 2
#     assert len(DateTimeIntYYYYMMDD.get_relations()) == 3
#
#
# def test_evolve_replace_datetime():
#     DateTimeIntYYYYMMDD = DateTime.evolve_type(
#         "int_yyyymmdd_r", relations_generator, replace=True
#     )
#     assert str(DateTimeIntYYYYMMDD) == "DateTime[int_yyyymmdd_r]"
#     assert str(DateTime) == "DateTime"
#     assert len(DateTime.get_relations()) == 2
#     assert len(DateTimeIntYYYYMMDD.get_relations()) == 1
